import argparse
import os

from colorama import init, Fore, Style
from pathlib import Path
# from IMDLBenCo.utils.paths import BencoPath
from IMDLBenCo.cli_funcs import cli_init, cli_guide, cli_data

def main():
    parser = argparse.ArgumentParser(description='Command line interface for IMDLBenCo')
    subparsers = parser.add_subparsers(dest='command', required=True)
    
    
    # init command
    parser_init = subparsers.add_parser('init', help='Initialize the environment')
    init_subparsers = parser_init.add_subparsers(dest='subcommand', required=False)
    
    # init base
    parser_init_base = init_subparsers.add_parser('base', help='Initialize the base environment')
    parser_init_base.set_defaults(subcommand='base')

    # init model_zoo
    parser_init_model_zoo = init_subparsers.add_parser('model_zoo', help='Initialize the model zoo')
    
    # init backbone
    parser_init_backbone = init_subparsers.add_parser('backbone', help='Initialize the backbone')
    
    
    # guide command
    parser_guide = subparsers.add_parser('guide', help='Guide for using the tool')
    
    # data command
    parser_data = subparsers.add_parser('data', help='Manage data')
    
    parser.add_argument('--config', type=str, help='Path to the configuration file')
    
    args = parser.parse_args()
    
    
    if args.command == 'init':
        if args.subcommand is None:
            args.subcommand = 'base'
        cli_init(args.config, subcommand=args.subcommand)
        
    elif args.command == 'guide':
        cli_guide(args.config)
    elif args.command == 'data':
        cli_data(args.config)

def train(config):
    print(f'Training with config: {config}')

def evaluate(config):
    print(f'Evaluating with config: {config}')

if __name__ == '__main__':
    main()



# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# Copyright (c) 2023 Image Processing Research Group of University Federico II of Naples ('GRIP-UNINA').
#
# All rights reserved.
# This work should only be used for nonprofit purposes.
#
# By downloading and/or using any of these files, you implicitly agree to all the
# terms of the license, as specified in the document LICENSE.txt
# (included in this package) and online at
# http://www.grip.unina.it/download/LICENSE_OPEN.txt

"""
Created in September 2022
@author: davide.cozzolino
"""

import torch
import torch.nn.functional as F

    
def weighted_statistics_pooling(x, log_w=None):
        b = x.shape[0]
        c = x.shape[1]
        x = x.view(b,c,-1)
        
        if log_w is None:
            log_w = torch.zeros((b,1,x.shape[-1]), device=x.device)
        else:
            assert log_w.shape[0]==b
            assert log_w.shape[1]==1
            log_w = log_w.view(b,1,-1)
        
            assert log_w.shape[-1]==x.shape[-1]
        
        log_w = F.log_softmax(log_w, dim=-1)
        x_min = -torch.logsumexp(log_w-x, dim=-1)
        x_max =  torch.logsumexp(log_w+x, dim=-1)
        
        w = torch.exp(log_w)
        x_avg = torch.sum(w*x  , dim=-1)
        x_msq = torch.sum(w*x*x, dim=-1)
        
        x = torch.cat((x_min, x_max, x_avg, x_msq), dim=1)
        
        return x


# TODO JZH
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from .abstract_class import AbstractEvaluator
import torch.distributed as dist
import os

class Image_Accuracy(AbstractEvaluator):
    def __init__(self, threshold=0.5) -> None:
        super().__init__()
        self.name = "image-level Accuracy"
        self.desc = "image-level Accuracy"
        self.threshold = threshold
        self.true_cnt = torch.tensor(0.0, dtype=torch.float64, device='cuda')
        self.cnt = torch.tensor(0.0, dtype=torch.float64, device='cuda')

    def batch_update(self, predict_label, label, *args, **kwargs):
        predict = (predict_label > self.threshold).float().cuda()
        self.true_cnt += torch.tensor(torch.sum(predict * label).item() + torch.sum((1 - predict) * (1 - label)).item(), dtype=torch.float64, device='cuda')
        self.cnt += torch.tensor(len(label), dtype=torch.float64, device='cuda')
        return None

    def epoch_update(self):
        t = torch.tensor([self.true_cnt, self.cnt], dtype=torch.float64, device='cuda')
        dist.barrier()
        dist.all_reduce(t, op=dist.ReduceOp.SUM)
        true_cnt = t[0].item()
        cnt = t[1].item()
        acc = true_cnt / cnt
        return acc
    
    def recovery(self):
        self.true_cnt = torch.tensor(0.0, dtype=torch.float64, device='cuda')
        self.cnt = torch.tensor(0.0, dtype=torch.float64, device='cuda')

class Pixel_Accuracy(AbstractEvaluator):
    def __init__(self,threshold = 0.5, mode="origin") -> None:
        super().__init__()
        self.name = "pixel-level Accuracy"
        self.desc = "pixel-level Accuracy"
        self.threshold = threshold
        #  mode : "origin, reverse, double"
        self.mode = mode

    def Cal_Confusion_Matrix(self, predict, mask, shape_mask):
        """compute local confusion matrix for a batch of predict and target masks
        Args:
            predict (_type_): _description_
            mask (_type_): _description_
            region (_type_): _description_
            
        Returns:
            TP, TN, FP, FN
        """
        threshold = self.threshold
        predict = (predict > threshold).float()
        if(shape_mask != None):
            TP = torch.sum(predict * mask * shape_mask, dim=(1, 2, 3))
            TN = torch.sum((1-predict) * (1-mask) * shape_mask, dim=(1, 2, 3))
            FP = torch.sum(predict * (1-mask) * shape_mask, dim=(1, 2, 3))
            FN = torch.sum((1-predict) * mask * shape_mask, dim=(1, 2, 3))
        else:
            TP = torch.sum(predict * mask, dim=(1, 2, 3))  
            TN = torch.sum((1-predict) * (1-mask), dim=(1, 2, 3)) 
            FP = torch.sum(predict * (1-mask), dim=(1, 2, 3)) 
            FN = torch.sum((1-predict) * mask, dim=(1, 2, 3))         
        return TP, TN, FP, FN

    def Cal_Reverse_Confusion_Matrix(self, predict, mask, shape_mask):
        """compute local confusion matrix for a batch of predict and target masks
        Args:
            predict (_type_): _description_
            mask (_type_): _description_
            region (_type_): _description_
            
        Returns:
            TP, TN, FP, FN
        """
        threshold = self.threshold
        predict = (predict > threshold).float()
        if(shape_mask != None):
            TP = torch.sum((1-predict) * mask * shape_mask, dim=(1, 2, 3))
            TN = torch.sum(predict * (1-mask) * shape_mask, dim=(1, 2, 3))
            FP = torch.sum((1-predict) * (1-mask) * shape_mask, dim=(1, 2, 3))
            FN = torch.sum(predict * mask * shape_mask, dim=(1, 2, 3))
        else:
            TP = torch.sum((1-predict) * mask, dim=(1, 2, 3))
            TN = torch.sum(predict * (1-mask), dim=(1, 2, 3))
            FP = torch.sum((1-predict) * (1-mask), dim=(1, 2, 3))
            FN = torch.sum(predict * mask, dim=(1, 2, 3))
        return TP, TN, FP, FN
    def batch_update(self, predict, mask, shape_mask=None, *args, **kwargs):
        if self.mode == "origin":
            TP, TN, FP, FN = self.Cal_Confusion_Matrix(predict, mask, shape_mask)
            ACC = (TP + TN)/(TP + TN + FP + FN)
        elif self.mode == "reverse":
            TP, TN, FP, FN = self.Cal_Reverse_Confusion_Matrix(predict, mask, shape_mask)
            ACC = (TP + TN)/(TP + TN + FP + FN)
        elif self.mode == "double":
            # TODO
            TP, TN, FP, FN = self.Cal_Confusion_Matrix(predict, mask, shape_mask)
            ACC = torch.max((TP + TN)/(TP + TN + FP + FN), (FP + FN)/(TP + TN + FP + FN))
        else:
            raise RuntimeError(f"Cal_ACC no mode name {self.mode}")
        return ACC
    def epoch_update(self):

        return None
    
    def recovery(self):
        return None


def test_origin_image_ACC():
    import os
    import torch
    import torch.distributed as dist
    
    # 初始化分布式环境
    dist.init_process_group(backend='nccl', init_method='env://')
    
    num_gpus = torch.cuda.device_count()
    local_rank = int(os.environ['LOCAL_RANK'])
    torch.cuda.set_device(local_rank)
    
    if dist.get_rank() == 0:
        print("number of GPUs", num_gpus)
    
    DATA_LEN = 200
    float_tensor = torch.rand(DATA_LEN * num_gpus).cuda(local_rank)  # 生成一个长度为 200*num_gpus 的浮点数 tensor
    int_tensor = torch.randint(0, 2, (DATA_LEN * num_gpus,)).cuda(local_rank)  # 生成一个包含 0 或 1 的整数 tensor
    
    evaluator = Image_Accuracy(threshold=0.5)
    dist.barrier()
    dist.broadcast(float_tensor, src=0)
    dist.broadcast(int_tensor, src=0)
    
    all_predicts = []
    all_labels = []
    
    idx = dist.get_rank() * DATA_LEN
    if dist.get_rank() != num_gpus - 1:
        predict_labels = float_tensor[idx: idx + DATA_LEN].cuda(local_rank)
        true_labels = int_tensor[idx: idx + DATA_LEN].cuda(local_rank)
    else:
        predict_labels = float_tensor[idx: idx + DATA_LEN - 50].cuda(local_rank)
        true_labels = int_tensor[idx: idx + DATA_LEN - 50].cuda(local_rank)
    
    if dist.get_rank() == 0:  # 只在 rank 0 进程中收集数据
        all_predicts = float_tensor[:-50].cpu().numpy()
        all_labels = int_tensor[:-50].cpu().numpy()
        all_pred = (all_predicts > 0.5).astype(float)
        acc = (sum(all_pred * all_labels) + sum((1 - all_pred) * (1 - all_labels))) / len(all_pred)
        print(f"Calculated Accuracy: {acc}")
    
    # 运行 batch_update 更新统计数据
    evaluator.batch_update(predict_labels, true_labels)
    
    # 模拟一个 epoch 结束，调用 epoch_update 来计算 Accuracy
    gpu_acc = evaluator.epoch_update()
    
    if dist.get_rank() == 0:
        print(f"ACC Score: {gpu_acc}")
    
    # 清理分布式环境
    dist.destroy_process_group()



def test_pixal_ACC():
    batch_size, channels, height, width = 4, 1, 10, 10
    predict = torch.rand(batch_size, channels, height, width)
    mask = torch.randint(0, 2, (batch_size, channels, height, width)).float()
    
    # 生成一个 shape_mask
    shape_mask = torch.randint(0, 2, (batch_size, channels, height, width)).float()
    # shape_mask = None
    acc = Pixel_Accuracy(mode="origin")
    acc_value_pytorch = acc.batch_update(predict, mask, shape_mask)
    print(acc_value_pytorch)

if __name__ == "__main__":
    # from sklearn.metrics import accuracy_score
    # print("test Image F1")
    # image_F1 = Image_Accuracy(threshold=0.5)
    # predict = torch.tensor([[0.9], [0.3], [0.4]])
    # label = torch.tensor([[1],[0],[1]])
    # print("my Acc:",image_F1(predict, label))

    # predict_binary = predict >= 0.5
    # # 将torch张量转换为numpy数组，因为sklearn的函数期望输入是numpy数组
    # predict_binary_np = predict_binary.numpy()
    # labels_np = label.numpy()

    # # 使用sklearn计算F1分数
    # acc = accuracy_score(labels_np, predict_binary_np)

    # print("Test Image Acc:", acc)
    # test_origin_image_ACC()
    test_pixal_ACC()


import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from .abstract_class import AbstractEvaluator
import torch.distributed as dist
import os
from sklearn.metrics import f1_score

class ImageF1(AbstractEvaluator):
    def __init__(self, threshold = 0.5) -> None:
        super().__init__()
        self.name = "image-level F1"
        self.desc = "image-level F1"
        self.threshold = threshold
        self.TP = 0
        self.TN = 0
        self.FP = 0
        self.FN = 0
        self.cnt = 0
    def batch_update(self, predict_label, label, *args, **kwargs):
        predict = (predict_label > self.threshold).float()
        self.TP += torch.sum(predict * label).item()
        self.TN += torch.sum((1-predict) * (1-label)).item()
        self.FP += torch.sum(predict * (1-label)).item()
        self.FN += torch.sum((1-predict) * label).item()
        self.cnt += len(predict_label)
        return None
    
    def epoch_update(self):
        t = torch.tensor([self.TP, self.TN, self.FP, self.FN, self.cnt],  dtype=torch.float64, device='cuda')
        dist.barrier()
        dist.all_reduce(t, op=dist.ReduceOp.SUM)
        TP = t[0].item()
        TN = t[1].item()
        FP = t[2].item()
        FN = t[3].item()
        cnt = t[4].item()
        precision = TP / ( TP +  FP + 1e-9)
        recall =  TP / ( TP +  FN + 1e-9)
        F1 = 2 * precision * recall / (precision + recall + 1e-9)
        # F1 = torch.mean(F1) # fuse the Batch dimension
        return F1
    def recovery(self):
        self.TP = 0
        self.TN = 0
        self.FP = 0
        self.FN = 0
        self.cnt = 0

            
class PixelF1(AbstractEvaluator):
    def __init__(self, threshold = 0.5, mode = "origin") -> None:
        super().__init__()
        self.name = "pixel-level F1"
        self.desc = "pixel-level F1"
        self.threshold = threshold
        self.image_num = 0
        #  mode : "origin, reverse, double"
        self.mode = mode

    def Cal_Confusion_Matrix(self, predict, mask, shape_mask):
        """compute local confusion matrix for a batch of predict and target masks
        Args:
            predict (_type_): _description_
            mask (_type_): _description_
            region (_type_): _description_
            
        Returns:
            TP, TN, FP, FN
        """
        threshold = self.threshold
        predict = (predict > threshold).float()
        if(shape_mask != None):
            TP = torch.sum(predict * mask * shape_mask, dim=(1, 2, 3))
            TN = torch.sum((1-predict) * (1-mask) * shape_mask, dim=(1, 2, 3))
            FP = torch.sum(predict * (1-mask) * shape_mask, dim=(1, 2, 3))
            FN = torch.sum((1-predict) * mask * shape_mask, dim=(1, 2, 3))
        else:
            TP = torch.sum(predict * mask, dim=(1, 2, 3))  
            TN = torch.sum((1-predict) * (1-mask), dim=(1, 2, 3)) 
            FP = torch.sum(predict * (1-mask), dim=(1, 2, 3)) 
            FN = torch.sum((1-predict) * mask, dim=(1, 2, 3))         
        return TP, TN, FP, FN

    def Cal_Reverse_Confusion_Matrix(self, predict, mask, shape_mask):
        """compute local confusion matrix for a batch of predict and target masks
        Args:
            predict (_type_): _description_
            mask (_type_): _description_
            region (_type_): _description_
            
        Returns:
            TP, TN, FP, FN
        """
        threshold = self.threshold
        predict = (predict > threshold).float()
        if(shape_mask != None):
            TP = torch.sum((1-predict) * mask * shape_mask, dim=(1, 2, 3))
            TN = torch.sum(predict * (1-mask) * shape_mask, dim=(1, 2, 3))
            FP = torch.sum((1-predict) * (1-mask) * shape_mask, dim=(1, 2, 3))
            FN = torch.sum(predict * mask * shape_mask, dim=(1, 2, 3))
        else:
            TP = torch.sum((1-predict) * mask, dim=(1, 2, 3))
            TN = torch.sum(predict * (1-mask), dim=(1, 2, 3))
            FP = torch.sum((1-predict) * (1-mask), dim=(1, 2, 3))
            FN = torch.sum(predict * mask, dim=(1, 2, 3))
        return TP, TN, FP, FN

    def Cal_F1(self, TP, TN, FP, FN):
        """_summary_

        Args:
            TP (_type_): _description_
            TN (_type_): _description_
            FP (_type_): _description_
            FN (_type_): _description_

        Returns:
            _type_: _description_
        """
        precision = TP / (TP + FP + 1e-8)
        recall = TP / (TP + FN + 1e-8)
        F1 = 2 * precision * recall / (precision + recall + 1e-8)
        # F1 = torch.mean(F1) # fuse the Batch dimension
        return F1

    def batch_update(self, predict, mask, shape_mask=None, *args, **kwargs): # 注意这里只有pixel-level需要的信息
        if self.mode == "origin":
            TP, TN, FP, FN = self.Cal_Confusion_Matrix(predict, mask, shape_mask)
            F1 = self.Cal_F1(TP, TN, FP, FN)
        elif self.mode == "reverse":
            TP, TN, FP, FN = self.Cal_Reverse_Confusion_Matrix(predict, mask, shape_mask)
            F1 = self.Cal_F1(TP, TN, FP, FN)
        elif self.mode == "double":
            # todo
            TP, TN, FP, FN = self.Cal_Confusion_Matrix(predict, mask, shape_mask)
            F1 = torch.max(self.Cal_F1(TP, TN, FP, FN), self.Cal_F1(FN, FP, TN, TP))
        else:
            raise RuntimeError(f"Cal_F1 no mode name {self.mode}")
        
        return F1
    def epoch_update(self):

        return None
    def recovery(self):
        self.image_num = 0


def test_origin_image_f1():
    # test imageF1
    # 初始化分布式环境
    dist.init_process_group(backend='nccl', init_method='env://')
    
    num_gpus = torch.cuda.device_count()
    if dist.get_rank() == 0:
        print("number of GPUS", num_gpus)
    
    local_rank = int(os.environ['LOCAL_RANK'])
    torch.cuda.set_device(local_rank)
    
    DATA_LEN = 200
    float_tensor = torch.rand( DATA_LEN * num_gpus).cuda(local_rank)  # 生成一个长度为 5 的浮点数 tensor

    # 生成一个包含 0 或 1 的整数 tensor
    int_tensor = torch.randint(0, 2, (DATA_LEN * num_gpus,)).cuda(local_rank)
    # print(float_tensor)
    # print(int_tensor)
    
    evaluator = ImageF1(threshold=0.5)
    dist.barrier()
    dist.broadcast(float_tensor, src=0)
    dist.broadcast(int_tensor, src=0)
    # 收集所有的预测标签和真实标签，用于之后的 sklearn 验证
    all_predicts = []
    all_labels = []
    
    
    if dist.get_rank() != num_gpus - 1:
        idx = dist.get_rank() * DATA_LEN
        predict_labels = float_tensor[idx: idx + DATA_LEN].cuda(local_rank)
        true_labels = int_tensor[idx: idx + DATA_LEN].cuda(local_rank)
    else:
        idx = dist.get_rank() * DATA_LEN
        predict_labels = float_tensor[idx: idx + DATA_LEN-50].cuda(local_rank)
        true_labels = int_tensor[idx: idx + DATA_LEN-50].cuda(local_rank)


    if dist.get_rank() == 0:  # 只在 rank 0 进程中收集数据
        all_predicts = float_tensor.cpu().numpy()
        all_labels= int_tensor.cpu().numpy()
        # print(all_labels)
            

    # 运行 batch_update 更新统计数据
    evaluator.batch_update(predict_labels, true_labels)
    
    # 模拟一个 epoch 结束，调用 epoch_update 来计算 F1 分数
    gpu_f1_score = evaluator.epoch_update()
    if(dist.get_rank() == 0):
        print(f"F1 Score: {gpu_f1_score}")
        # 使用 sklearn 计算 F1 分数
        # all_predicts = np.concatenate(all_predicts)
        # all_labels = np.concatenate(all_labels)
        sklearn_f1 = f1_score(all_labels[:-50], (all_predicts[:-50] > 0.5).astype(int), average='binary')
        print(f"F1 Score (sklearn): {sklearn_f1}", "cnt = ", len(all_labels[:-50]))
    

    # 清理分布式环境
    dist.destroy_process_group()


            

if __name__ == "__main__":
    test_origin_image_f1()


import torch.nn as nn
"""
下面这个类很重要，可以自动管理在多卡之间的一个具体数值变量的reduce（显卡之间归并数据）
可以用于实现算法的参考。
"""
# from training.utils.misc import MetricLogger

"""
改变这个接口的主要目的是image-level的指标和pixel-level的指标的计算方式不同
"""
class AbstractEvaluator(object): # 想了想没必要用nn.module 反而可能会引起一些其他的问题，比如梯度追踪，或者Parameter的追踪等等问题？？？（我有点困不是很确定）
    def __init__(self) -> None:
        self.name = None
        self.desc = None
        self.threshold = None
    def batch_update(self, predict, pred_label, mask, shape_mask=None, *args, **kwargs):
        """
        本函数在每个batch结尾update。
        """
        raise NotImplementedError
    def epoch_update(self):
        """
        理论上这个时候没有新的数据了，所以没有输入参数。
        
        功能：在显卡之间收集所有在整个epoch内统计的指标，然后返回最终期望的信息。
        """
        raise NotImplementedError
    def recovery(self):
        raise NotImplementedError




    



import os
import cv2
import numpy as np
from PIL import Image

import torch
from torch.utils.data import Dataset, DataLoader

from .utils import pil_loader, denormalize

from ..transforms import get_albu_transforms, EdgeMaskGenerator

from IMDLBenCo.registry import DATASETS

@DATASETS.register_module()
class AbstractDataset(Dataset):
    def _init_dataset_path(self, path):
        tp_path = None # Tampered image
        gt_path = None # Ground truth
        
        raise NotImplementedError # abstract dataset!
    
        return tp_path, gt_path, labels
        
    def __init__(self, path, 
                is_padding = False,
                is_resizing = False,
                output_size = (1024, 1024),
                common_transforms = None, 
                edge_width = None,
                img_loader = pil_loader,
                post_funcs = None
                ) -> None:
        super().__init__()
        self.tp_path, self.gt_path = self._init_dataset_path(path)
        
        if is_padding == True and is_resizing == True:
            raise AttributeError("is_padding and is_resizing can not be True at the same time")
        if is_padding == False and is_resizing == False:
            raise AttributeError("is_padding and is_resizing can not be False at the same time")

        # Padding or Resizing
        self.post_transform = None
        if is_padding == True:
            self.post_transform = get_albu_transforms(type_ = "pad", output_size = output_size)
        if is_resizing == True:
            self.post_transform = get_albu_transforms(type_ = "resize", output_size = output_size)
        self.is_padding = is_padding
        self.is_resizing = is_resizing
        
        self.output_size = output_size
        
        # Common augmentations for augumentation
        self.common_transforms = common_transforms
        # Edge mask generator        
        self.edge_mask_generator = None if edge_width is None else EdgeMaskGenerator(edge_width)

        self.img_loader = img_loader
        self.post_funcs = post_funcs
        
    def __getitem__(self, index):

        data_dict = dict()
        
        tp_path = self.tp_path[index]
        gt_path = self.gt_path[index]
        
        # pil_loader or jpeg_loader
        tp_img = self.img_loader(tp_path)
        # shape, here is PIL Image
        tp_shape = tp_img.size
        
        # if "negative" then gt is a image with all 0
        if gt_path != "Negative":
            gt_img = self.img_loader(gt_path)
            gt_shape = gt_img.size
            label = 1
        else:
            temp = np.array(tp_img)
            gt_img = np.zeros((temp.shape[0], temp.shape[1], 3))
            gt_shape = (temp.shape[1], temp.shape[0])
            label = 0
            
        assert tp_shape == gt_shape, "tp and gt image shape must be the same, but got {} and {}".format(tp_shape, gt_shape)
        
        tp_img = np.array(tp_img) # H W C
        gt_img = np.array(gt_img) # H W C
        
        # Do augmentations
        if self.common_transforms != None:
            res_dict = self.common_transforms(image = tp_img, mask = gt_img)
            tp_img = res_dict['image']
            gt_img = res_dict['mask']
            
        # redefine the shape, here is np.array
        tp_shape = tp_img.shape[0:2]  # H, W, 3 去掉最后一个3
        
        gt_img =  (np.mean(gt_img, axis = 2, keepdims = True)  > 127.5 ) * 1.0 # fuse the 3 channels to 1 channel, and make it binary(0 or 1)
        gt_img =  gt_img.transpose(2,0,1)[0] # H W C -> C H W -> H W
        masks_list = [gt_img]
        
        # if need to generate broaden edge mask
        if self.edge_mask_generator != None: 
            gt_img_edge = self.edge_mask_generator(gt_img)[0][0] # B C H W -> H W
            masks_list.append(gt_img_edge) # albumentation interface
        else:
            pass
            
        # Do post-transform (paddings or resizing)    
        res_dict = self.post_transform(image = tp_img, masks = masks_list)
        
        tp_img = res_dict['image']
        gt_img = res_dict['masks'][0].unsqueeze(0) # H W -> 1 H W \
        
        
        if self.edge_mask_generator != None:
            gt_img_edge = res_dict['masks'][1].unsqueeze(0) # H W -> 1 H W  

            # =========output=====================
            data_dict['edge_mask'] = gt_img_edge
            # ====================================
        # name of the image (mainly for testing)
        basename = os.path.basename(tp_path)

        # =========output=====================
        data_dict['image'] = tp_img
        data_dict['mask'] = gt_img
        data_dict['label'] = label
       
        # 如果经过resize
        if self.is_resizing:
            tp_shape = self.output_size
            
        # 这里如果是（256， 384） 那么对应的图像是一个横着的 长的方块
        data_dict['shape'] = torch.tensor(tp_shape) # (H, W) 经过data loader后会变成三维矩阵，第0维是batch_index
        data_dict['name'] = basename
        
        # 如果padding则需要单独return一个shape_mask
        if self.is_padding:
            shape_mask = torch.zeros_like(gt_img)
            shape_mask[:, :tp_shape[0], :tp_shape[1]] = 1
            data_dict['shape_mask'] = shape_mask
        # ====================================
        # Post processing with callback functions on data_dict
        if self.post_funcs == None:
            pass    # Do nothing
        elif isinstance(self.post_funcs, list):
            # 如果是列表，循环调用列表中的每个回调函数
            for func in self.post_funcs:
                if callable(func):
                    func(data_dict)
                else:
                    raise NotImplementedError(f"Element {func} in list is not callable")
        elif callable(self.post_funcs):
            # 如果是单个回调函数，直接调用
            self.post_funcs(data_dict)
        else:
            # 其他类型抛出 NotImplementedError
            raise NotImplementedError(f"Unsupported type: {type(self.post_funcs)}")
        # ====================================

        return data_dict
        
    def __len__(self):
        return len(self.tp_path)



from .iml_transforms import get_albu_transforms, RandomCopyMove, RandomInpainting
from .edge_mask_generator import EdgeMaskGenerator

__all__ = ['get_albu_transforms', 'RandomCopyMove', 'RandomInpainting', 'EdgeMaskGenerator']


import json
import os


from ..registry import DATASETS
from .abstract_dataset import AbstractDataset
    
@DATASETS.register_module()
class ManiDataset(AbstractDataset):
    def _init_dataset_path(self, path):
        path = path
        tp_dir = os.path.join(path, 'Tp')
        gt_dir = os.path.join(path, 'Gt')
        tp_list = os.listdir(tp_dir)
        gt_list = os.listdir(gt_dir)
        # Use sort mathod to keep order, to make sure the order is the same as the order in the tp_list and gt_list
        tp_list.sort()
        gt_list.sort()
        t_tp_list = [os.path.join(path, 'Tp', tp_list[index]) for index in range(len(tp_list))]
        t_gt_list = [os.path.join(path, 'Gt', gt_list[index]) for index in range(len(gt_list))]
        return t_tp_list, t_gt_list
    
@DATASETS.register_module()
class JsonDataset(AbstractDataset):
    """ init from a json file, which contains all the images path
        file is organized as:
            [
                ["./Tp/6.jpg", "./Gt/6.jpg"],
                ["./Tp/7.jpg", "./Gt/7.jpg"],
                ["./Tp/8.jpg", "Negative"],
                ......
            ]
        if path is "Neagative" then the image is negative sample, which means ground truths is a totally black image, and its label == 0.
        
    Args:
        path (_type_): _description_
        transform_albu (_type_, optional): _description_. Defaults to None.
        mask_edge_generator (_type_, optional): _description_. Defaults to None.
        if_return_shape
    """
    def _init_dataset_path(self, path):
        images = json.load(open(path, 'r'))
        tp_list = []
        gt_list = []
        for record in images:
            if os.path.isfile(record[0]):
                tp_list.append(record[0])
                gt_list.append(record[1])
            else: 
                raise TypeError("Not a file in Json Dataset Error. Try other dataset")
        return tp_list, gt_list




import torch
import torch.nn as nn
import torch.fft
import math
class DCTExtractor(nn.Module):
    def __init__(self, alpha=0.1):
        super(DCTExtractor, self).__init__()
        if alpha <= 0 or alpha >= 1:
            raise ValueError("alpha must be between 0 and 1 (exclusive)")
        self.alpha = alpha
        self.dct_matrix_h = None
        self.dct_matrix_w = None

    def create_dct_matrix(self, N):
        n = torch.arange(N, dtype=torch.float32).reshape((1, N))
        k = torch.arange(N, dtype=torch.float32).reshape((N, 1))
        dct_matrix = torch.sqrt(torch.tensor(2.0 / N)) * torch.cos(math.pi * k * (2 * n + 1) / (2 * N))
        dct_matrix[0, :] = 1 / math.sqrt(N)
        return dct_matrix

    def dct_2d(self, x):
        H, W = x.size(-2), x.size(-1)
        if self.dct_matrix_h is None or self.dct_matrix_h.size(0) != H:
            self.dct_matrix_h = self.create_dct_matrix(H).to(x.device)
        if self.dct_matrix_w is None or self.dct_matrix_w.size(0) != W:
            self.dct_matrix_w = self.create_dct_matrix(W).to(x.device)
        
        return torch.matmul(self.dct_matrix_h, torch.matmul(x, self.dct_matrix_w.t()))

    def idct_2d(self, x):
        H, W = x.size(-2), x.size(-1)
        if self.dct_matrix_h is None or self.dct_matrix_h.size(0) != H:
            self.dct_matrix_h = self.create_dct_matrix(H).to(x.device)
        if self.dct_matrix_w is None or self.dct_matrix_w.size(0) != W:
            self.dct_matrix_w = self.create_dct_matrix(W).to(x.device)
        
        return torch.matmul(self.dct_matrix_h.t(), torch.matmul(x, self.dct_matrix_w))

    def high_pass_filter(self, x, alpha):
        h, w = x.shape[-2:]
        mask = torch.ones(h, w, device=x.device)
        alpha_h, alpha_w = int(alpha * h), int(alpha * w)
        mask[:alpha_h, :alpha_w] = 0
        return x * mask

    def forward(self, x):
        xq = self.dct_2d(x)
        xq_high = self.high_pass_filter(xq, self.alpha)
        xh = self.idct_2d(xq_high)
        B = xh.shape[0]
        min_vals = xh.reshape(B, -1).min(dim=1, keepdim=True).values.view(B, 1, 1, 1)
        max_vals = xh.reshape(B, -1).max(dim=1, keepdim=True).values.view(B, 1, 1, 1)
        xh = (xh - min_vals) / (max_vals - min_vals)
        return xh

class FFTExtractor(nn.Module):
    def __init__(self, alpha=0.1):
        super(FFTExtractor, self).__init__()
        if alpha <= 0 or alpha >= 1:
            raise ValueError("alpha must be between 0 and 1 (exclusive)")
        self.alpha = alpha
        
    def fft_2d(self, x):
        return torch.fft.fftshift(torch.fft.fft2(x))

    def ifft_2d(self, x):
        return torch.fft.ifft2(torch.fft.ifftshift(x))

    def high_pass_filter(self, x, alpha):
        h, w = x.shape[-2:]
        mask = torch.ones(h, w, device=x.device)
        alpha_h, alpha_w = int(alpha * h) // 2, int(alpha * w) // 2
        mask_center_h = h // 2
        mask_center_w = w // 2
        mask[mask_center_h - alpha_h:mask_center_h + alpha_h, mask_center_w - alpha_w:mask_center_w + alpha_w] = 0
        mask = mask.expand_as(x)
        return x * mask

    def forward(self, x):
        xq = self.fft_2d(x)
        xq_high = self.high_pass_filter(xq, self.alpha)
        xh = self.ifft_2d(xq_high)
        xh = xh.real
        B = xh.shape[0]
        min_vals = xh.reshape(B, -1).min(dim=1, keepdim=True).values.reshape(B, 1, 1, 1)
        max_vals = xh.reshape(B, -1).max(dim=1, keepdim=True).values.reshape(B, 1, 1, 1)
        xh = (xh - min_vals) / (max_vals - min_vals)
        return xh




import os
from colorama import init, Fore, Style
from IMDLBenCo.utils.paths import BencoPath
from .copy import copy_files, copy_file 

def cli_guide(config):
    print("This is guide cli.")
    # TODO


import os
import re
from colorama import init, Fore, Style
from IMDLBenCo.utils.paths import BencoPath
from .copy import copy_files, copy_file 

def _copy_train_scripts():
    current_dir = os.getcwd()
    # Copy train scripts
    target_dir = os.path.join(current_dir)
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)
    copy_files(BencoPath.get_templates_dir(), target_dir)
    
    
def _copy_dataset_json():
    current_dir = os.getcwd()
    # Copy train scripts
    target_dir = os.path.join(current_dir)
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)
    copy_files(BencoPath.get_dataset_json_dir(), target_dir)
    
def _copy_init_base_files():
    current_dir = os.getcwd()
    # Copy train scripts
    target_dir = os.path.join(current_dir)
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)
    copy_files(BencoPath.get_init_base_dir(), target_dir)
    
def _copy_demo_runs():
    # Copy demo runs
    current_dir = os.getcwd()
    target_dir = os.path.join(current_dir, 'runs')
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)
    copy_files(BencoPath.get_model_zoo_runs_dir(), target_dir)
    
def _copy_demo_configs():
    # Copy demo configs
    current_dir = os.getcwd()
    target_dir = os.path.join(current_dir, 'configs')
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)
    copy_files(BencoPath.get_model_zoo_configs_dir(), target_dir)
    
# hejack train.py test.py test_robust.py with self defined model
def _inject_after_last_import(path, inject):
    # 读取文件内容
    with open(path, 'r') as file:
        content = file.readlines()
    
    base_file_name = os.path.basename(path)
    # 检查是否已经存在指定的导入语句
    if any(inject in line for line in content):
        print(f"  The specified import statement already exists in {base_file_name}.")
        return
    
    # 正则表达式匹配 from ... import ... 行
    pattern = re.compile(r'^\s*from\s+\S+\s+import\s+(?!.*\()(\S+(,\s*\S+)*)\s*$')
    
    # 找到最后一个匹配的行的索引
    last_import_index = -1
    for i, line in enumerate(content):
        if pattern.match(line):
            last_import_index = i
    
    # 如果找到匹配的行，则在其后插入新的内容
    if last_import_index != -1:
        content.insert(last_import_index + 1, inject + '\n')
    
    # re-save the file
    with open(path, 'w') as file:
        file.writelines(content)
    print(f"  Injected '{inject}' into {path}.")

def cli_init(config, subcommand):
    print(f'{Fore.GREEN}Initializing in current working directory...{Style.RESET_ALL}')
    
    # base initialize that only contain default scripts
    if subcommand == "base":
        _copy_train_scripts()
        _copy_init_base_files()
        _copy_dataset_json()
        for name in ['train.py', 'test.py', 'test_robust.py']:
            injected_str = "from mymodel import MyModel"
            current_dir = os.getcwd()
            target_dir = os.path.join(current_dir, name)
            _inject_after_last_import(target_dir, injected_str)
        
    if subcommand == "model_zoo":
        _copy_train_scripts()
        _copy_demo_runs() 
        _copy_demo_configs()
        _copy_dataset_json()
    # base initialize that only contain default scripts
    if subcommand == "backbone":
        _copy_train_scripts()
        _copy_demo_runs() 
        _copy_demo_configs()
        _copy_dataset_json()
    print(f'{Fore.GREEN}Successfully initialized IMDLBenCo scripts.{Style.RESET_ALL}')


import torch
import torch.nn as nn
import torch.nn.functional as F

class PixelAttention(nn.Module):
    def __init__(self, in_channels=3, kernel_range=[3, 3], shift=1, use_bn=False, use_res=False):
        super(PixelAttention, self).__init__()
        self.in_channels = in_channels
        self.kernel_range = kernel_range
        self.shift = shift
        self.use_bn = use_bn
        self.use_res = use_res

        n_p = kernel_range[0] * kernel_range[1]
        self.k_p = nn.Conv2d(in_channels, in_channels * n_p, kernel_size=1)
        self.v_p = nn.Conv2d(in_channels, in_channels * n_p, kernel_size=1)
        self.q_p = nn.Conv2d(in_channels, in_channels, kernel_size=1)
        
        self.ff1 = nn.Conv2d(in_channels, in_channels, kernel_size=3, padding=1)
        self.ff2 = nn.Conv2d(in_channels, 2 * in_channels, kernel_size=3, padding=1)
        self.ff3 = nn.Conv2d(2 * in_channels, in_channels, kernel_size=3, padding=1)
        
        if self.use_bn:
            self.bn1 = nn.BatchNorm2d(in_channels)
            self.bn2 = nn.BatchNorm2d(in_channels)

    def forward(self, x):
        h_half = self.kernel_range[0] // 2
        w_half = self.kernel_range[1] // 2
        s = x.shape
        D = s[1]
        # print("D:", D)
        x_k = self.k_p(x)
        x_v = self.v_p(x)
        x_q = self.q_p(x)
        # print("x_k.shape: ", x_k.shape)
        paddings = (w_half * self.shift, w_half * self.shift, h_half * self.shift, h_half * self.shift)
        # print("paddings: ", paddings)
        x_k = F.pad(x_k, paddings, "constant", 0)
        x_v = F.pad(x_v, paddings, "constant", 0)
        mask_x = torch.ones(s[0], 1, s[2], s[3], device=x.device)
        # print("x_k.shape: ", x_k.shape)
        # print("mask_x.shape: ", mask_x.shape)
        mask_pad = F.pad(mask_x, paddings, "constant", 0)
        
        k_ls = []
        v_ls = []
        masks = []
        
        c_x, c_y = h_half * self.shift, w_half * self.shift
        layer = 0
        # print("s[2]: ", s[2])
        # print("s[3]: ", s[3])
        for i in range(-h_half, h_half + 1):
            for j in range(-w_half, w_half + 1):
                
                k_t = x_k[:, layer*D:(layer+1)*D, c_x + i * self.shift:c_x + i * self.shift + s[2], c_y + j * self.shift:c_y + j * self.shift + s[3]]
                k_ls.append(k_t)
                v_t = x_v[:, layer*D:(layer+1)*D, c_x + i * self.shift:c_x + i * self.shift + s[2], c_y + j * self.shift:c_y + j * self.shift + s[3]]
                v_ls.append(v_t)
                # print("mask_pad.shape:", mask_pad.shape)
                _m = mask_pad[:, :, c_x + i * self.shift:c_x + i * self.shift + s[2], c_y + j * self.shift:c_y + j * self.shift + s[3]]
                # print("v_t: ", v_t.shape)
                # print("_m: ", _m.shape)
                masks.append(_m)
                layer += 1
        # print("mask: ", masks[0].shape)
        m_stack = torch.stack(masks, dim=1)
        m_vec = m_stack.view(s[0] * s[2] * s[3], self.kernel_range[0] * self.kernel_range[1], 1)
        # print("len(k_ls): ", len(k_ls))
        # print("k_ls[0]: ", k_ls[0].shape)
        k_stack = torch.stack(k_ls, dim=1)
        v_stack = torch.stack(v_ls, dim=1)
        # print("k_stack: ", k_stack.shape)
        k = k_stack.view(s[0] * s[2] * s[3] , self.kernel_range[0] * self.kernel_range[1], D)
        v = v_stack.view(s[0] * s[2] * s[3] , self.kernel_range[0] * self.kernel_range[1], D)
        q = x_q.view(s[0] * s[2] * s[3], 1, D)
        
        alpha = F.softmax(torch.matmul(k, q.transpose(-1, -2)) * m_vec / 8, dim=1)
        __res = torch.matmul(alpha.transpose(-1, -2), v)
        _res = __res.view(s[0], D, s[2], s[3])
        
        if self.use_res:
            t = x + _res
        else:
            t = _res
        
        if self.use_bn:
            t = self.bn1(t)
        
        _t = t
        t = F.relu(self.ff1(t))
        t = F.relu(self.ff2(t))
        t = F.relu(self.ff3(t))
        
        if self.use_res:
            t = _t + t
        
        if self.use_bn:
            res = self.bn2(t)
        else:
            res = t
        
        return res

# 测试 PixelAttention 类
if __name__ == "__main__":
    input_tensor = torch.randn(1, 3, 64, 64)  # (batch_size, in_channels, height, width)
    pixel_attention = PixelAttention(3)
    output = pixel_attention(input_tensor)
    print("output:", output.shape)



from setuptools import setup, find_packages

from IMDLBenCo.version import __version__


"""
For documentations, see here:
https://github.com/pypa/sampleproject/blob/db5806e0a3204034c51b1c00dde7d5eb3fa2532e/setup.py

在工作路径下运行下列指令：
pip install -e .  
即可实现本地安装本包，并在更新文件时自动更新相应内容，便于调试开发。
"""

def readme():
    with open('README.md', encoding='utf-8') as f:
        content = f.read()
    return content

def requirements():
    with open('requirements.txt') as f:
        requirements = f.read().splitlines()
    return requirements

# exit(0)

setup(
    name='IMDLBenCo',
    version=__version__,
    description="A comprehensive benchmark and code base for Image manipulation and localization.",
    long_description=readme(),
    long_description_content_type='text/markdown',
    url="https://github.com/scu-zjz/IMDLBenCo",
    author="Xiaochen Ma",   # Optional
    author_email="xiaochen.ma.cs@gmail.com", # Optional
    

    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",

        "Topic :: Security",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        
        # Pick your license as you wish
        "License :: Free For Educational Use",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate you support Python 3. These classifiers are *not*
        # checked by 'pip install'. See instead 'python_requires' below.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords=[
        "AI",
        "artificial intelligence",
        "image forensics",
        "image manipulation localization",
        "image manipulation detection"
        ],  # Optional
    
    packages=find_packages(),
    python_requires=">=3.7, <4",
    

    include_package_data=True, #
    install_requires=requirements(),
    license='CC-BY-4.0',
    entry_points={
        'console_scripts': [
            'benco=IMDLBenCo.cli:main',  # cli是你的命令行脚本的模块，main是入口函数
        ],
    },
    project_urls={  # Optional
        "Github": "https://github.com/scu-zjz/IMDLBenCo/",
        "Documentation":"https://github.com/scu-zjz/IMDLBenCo-doc",
        "Bug Reports": "https://github.com/scu-zjz/IMDLBenCo/issues",
    },
)




import torch
import torch.nn as nn
import torch.nn.functional as F
import matplotlib.pyplot as plt
from sklearn.metrics import roc_auc_score,roc_curve
import numpy as np
from torch.utils.data import DataLoader
def genertate_region_mask(masks ,batch_shape:torch.Tensor):
    """generate B 1 H W meaningful-region-mask for a batch of masks

    Args:
        batch_shape (_type_): list of tensor, e.g. [ tensor(384, 256), tensor(256, 384), ....]
    """
    meaningful_mask = torch.zeros_like(masks)
    for idx, shape in enumerate(batch_shape):
        # shape = shape.tolist()
        meaningful_mask[idx, :, :shape[0], :shape[1]] = 1 # TODO 可能会出傻逼问题（长宽翻转）的地方 一定要做测试
    return meaningful_mask

def cal_confusion_matrix(predict, target, region_mask, threshold=0.5):
    """compute local confusion matrix for a batch of predict and target masks
    Args:
        predict (_type_): _description_
        target (_type_): _description_
        region (_type_): _description_
        
    Returns:
        TP, TN, FP, FN
    """
    predict = (predict > threshold).float()
    TP = torch.sum(predict * target * region_mask, dim=(1, 2, 3))
    TN = torch.sum((1-predict) * (1-target) * region_mask, dim=(1, 2, 3))
    FP = torch.sum(predict * (1-target) * region_mask, dim=(1, 2, 3))
    FN = torch.sum((1-predict) * target * region_mask, dim=(1, 2, 3))
    return TP, TN, FP, FN

def cal_F1(TP, TN, FP, FN):
    """_summary_

    Args:
        TP (_type_): _description_
        TN (_type_): _description_
        FP (_type_): _description_
        FN (_type_): _description_

    Returns:
        _type_: _description_
    """
    precision = TP / (TP + FP + 1e-8)
    recall = TP / (TP + FN + 1e-8)
    F1 = 2 * precision * recall / (precision + recall + 1e-8)
    # F1 = torch.mean(F1) # fuse the Batch dimension
    return F1
    
if __name__ == "__main__":
    pass


import torch
import torch.nn as nn
from typing import Union, Tuple
from torch.utils.data import Dataset
from .grad_cam_hack import GradCAM
import numpy as np
from pytorch_grad_cam.utils.image import show_cam_on_image
from ...datasets.utils import denormalize
import os
from PIL import Image
from tqdm import tqdm


class OutputWrapper(torch.nn.Module):
    def __init__(self, model): 
        super(OutputWrapper, self).__init__()
        self.model = model
        
    def forward(self, x):
        return self.model(**x)["pred_mask"]


class SemanticSegmentationTarget:
    def __init__(self, mask):
        self.mask = torch.squeeze(mask, 0)
        
    def __call__(self, model_output):
        return (model_output * self.mask).sum()


def grad_camera_visualize(model: nn.Module,
                        image: Union[Tuple[str, str], Dataset],
                        target_layers: list,
                        output_path: str) -> None:
    """
    visualize
    """
    if not isinstance(image, Dataset):
        # TODO
        pass 
    sampler = torch.utils.data.SequentialSampler(image)
    data_loader = torch.utils.data.DataLoader(
        image, 
        sampler=sampler,
        batch_size=1,
        num_workers=1,
        pin_memory=False,
        drop_last=False,
    )
    model = OutputWrapper(model=model)
    for data in tqdm(data_loader):
        for key, value in data.items():
            if isinstance(value, torch.Tensor):
                data[key] = value.cuda()
        input_tensor = data['image']
        cam = GradCAM(model=model, target_layers=target_layers)
        rgb_img = np.float32(torch.squeeze(denormalize(input_tensor), 0).permute(1, 2, 0).cpu())
        targets = [SemanticSegmentationTarget(data['mask'])]
        grayscale_cam = cam(data, input_tensor=input_tensor, targets=targets)[0, :]
        # TODO fix the reshape bug
        # target_shape = tuple(data['shape'].cpu().numpy().squeeze())
        # rgb_img_resized = np.array(Image.fromarray(rgb_img.astype('uint8')).resize(target_shape, Image.BILINEAR))
        # grayscale_cam_resized = np.array(Image.fromarray(grayscale_cam).resize(target_shape, Image.BILINEAR))
        # pdb.set_trace()
        cam_image = show_cam_on_image(rgb_img, grayscale_cam, use_rgb=True)
        Image.fromarray(cam_image).save(os.path.join(output_path, data['name'][0].split('.')[0] + '.jpg'))